#include "LoRa.h"
#include "esp_sleep.h"
#include "Arduino.h"
#include "CRC32.h"            // Checksum methods
#include "LangloLoRa.h"       // LoRa configuration parameters
#include "LangloPackets.h"    // Packet structures
#include "nodeS1config.h"     // Specific Node configuration parameters

// The counter needs to be stored in RTC memory to be preserved across deep sleep
RTC_DATA_ATTR int counter = 0;

// Rainfall conversion factor (?)
// const float LSB = 0.002761647; // per manual calibration
const float LSB = 0.02761647; // per manual calibration (when using int in 1/10mm)
int inputRainfall = 0;
int adc0 = 0;

// WIFI_LoRa_32 ports

// GPIO5 — SX1278’s SCK
// GPIO19 — SX1278’s MISO
// GPIO27 — SX1278’s MOSI
// GPIO18 — SX1278’s CS
// GPIO14 — SX1278’s RESET
// GPIO26 — SX1278’s IRQ(Interrupt Request)

#define CS 18
#define RST 14
#define DI0 26

dataPacket outgoingPacket = {0};
rainfallPayload packetPayload = {0};

void setup() {
  LoRa.setPins(CS,RST,DI0);
  if (!LoRa.begin(BAND, PABOOST)) {
    while (true);
  }
  LoRa.setSpreadingFactor(SpreadingFactor);
  LoRa.setSignalBandwidth(SignalBandwidth);
  LoRa.setPreambleLength(PreambleLength);
  LoRa.setCodingRate4(CodingRateDenominator);
  LoRa.setSyncWord(SyncWord);
  LoRa.disableCrc();
  delay(1000);

// Read the rainfall from ADC1_0 (Pin 36)
  adc0 = analogRead(36);

//  inputRainfall = adc0 * LSB;
  inputRainfall = 155; // Dummy value (in 1/10 mm) just so taht there's somethign other than 0 to report

  String rainfallString = String(inputRainfall/10, 1);

//  Serial.print("[Loop] Sending packet");
  
  packetPayload.recorded.rainfall = inputRainfall;

  sendMessage(counter, rainfallTypePacket, packetPayload.payloadByte);
  counter++;                 // Increment message count

  esp_sleep_enable_timer_wakeup(TIME_TO_SLEEP * uS_TO_S_FACTOR);
//  esp_sleep_pd_config(ESP_PD_DOMAIN_MAX, ESP_PD_OPTION_OFF);
//  esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_PERIPH, ESP_PD_OPTION_OFF);
//  esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_SLOW_MEM, ESP_PD_OPTION_OFF);
//  esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_FAST_MEM, ESP_PD_OPTION_OFF);
  esp_deep_sleep_start();

}

void loop() {
// There's nothing here because the processor and memory are effectively reset after coming out of deep sleep
}

void sendMessage(int counter, uint8_t packetType, uint8_t payload[]) {

  outgoingPacket.content.destinationMAC = gatewayMAC;
  outgoingPacket.content.sourceMAC = myMAC;
  outgoingPacket.content.sequenceNumber = (uint16_t) counter % 65536;
  outgoingPacket.content.type = packetType;
  outgoingPacket.content.byteCount = (uint8_t) bytesIn(packetType);
  for (int i = 0; i < outgoingPacket.content.byteCount; i++) {
    outgoingPacket.content.payload[i] = payload[i];
  }

// Finally, calculate the checksum

  outgoingPacket.content.checksum = generateChecksum(outgoingPacket);
  
  LoRa.beginPacket();             // start packet
  int totalByteCount = LGO_HeaderSize + outgoingPacket.content.byteCount;
  for (int i = 0; i < totalByteCount; i++) {
    LoRa.write(outgoingPacket.packetByte[i]);
  }
  LoRa.endPacket();             // Finish packet and send it
}

uint32_t generateChecksum(dataPacket& outgoingPacket) {
  
//  Generate the checksum on the outgoing packet

  CRC32 crc;
  int checksumStart = LGO_HeaderSize - 4;
  int checksumEnd = outgoingPacket.content.byteCount + 4;
  for (int i = checksumStart; i < checksumEnd; i++) {
    crc.update(outgoingPacket.packetByte[i]);
  }
  uint32_t packetChecksum = crc.finalize();
  return packetChecksum;
}

int bytesIn(uint8_t packetType) {
  int result;
  switch (packetType) {
    case powerTypePacket:
      result = powerPayloadBytes;
      break;
    case tankTypePacket:
      result = tankPayloadBytes;
      break;
    case pumpTypePacket:
      result = pumpPayloadBytes;
      break;
    case weatherTypePacket:
      result = weatherPayloadBytes;
      break;
    case atmosphereTypePacket:
      result = atmospherePayloadBytes;
      break;
    case rainfallTypePacket:
      result = rainfallPayloadBytes;
      break;
    case windTypePacket:
      result = windPayloadBytes;
      break;
    case sprinklerTypePacket:
      result = sprinklerPayloadBytes;
      break;
    default:
      result = defaultPayloadBytes;
  }
  return result;
}
